/*
 * By Z. Dodds,
 *    adapted from code by J Hodas with
 *    important improvements, e.g., line/word input
 *    is now compatable, one can get chars, etc.
 *      (prompting and error-catching) by Chris Erickson HMC '06
 *    Thanks, Chris!
 */

package H;
 
import H.*;
import java.io.*;
import java.lang.*;

public class HMCInput 
{
  private BufferedReader r;
  private StreamTokenizer theStream;
  private boolean good;
  private boolean tokensWaiting;
  private String lastLineRead;       // this holds the text yet to be processed

  /** 
   * Takes an InputStream and builds an HMCInput stream from it. 
   * 
   * @param I is an InputStream from which the formatted input stream
   * is built.
   */
    
  public HMCInput(InputStream I)
  {
      r = new BufferedReader(new InputStreamReader(I));
      theStream = new StreamTokenizer(r);

      theStream.resetSyntax();
      theStream.wordChars(33, 255);
      theStream.whitespaceChars(0, ' ');
      theStream.eolIsSignificant(true);
      
      good = true;
      tokensWaiting = false;
      lastLineRead = "";
  }


  /**
   * @return Returns true if the input stream is
   * in a good state (i.e. no input errors have occured, or the last error 
   * condition has been cleared.
   */

  public boolean good() 
  {
      return good; 
  }


  /**
   * @return Returns true if the stream has reached End Of File.
   */

  public boolean EOF() 
  { 
      return (theStream.ttype == StreamTokenizer.TT_EOF);
  }
  
  /*
   * this method removes initial whitespace from a string
   */
  String removeWhiteSpace(String s)
  {
    int startindex = 0;
    int len = s.length();
    
    while (startindex < len)
    {
      // anything less than ' ' is a space...
      if (s.charAt(startindex) <= ' ')   
      {
        ++startindex;  
      }
      else
        break;
    }
    
    String ret = "";
    
    try
    {
        ret = s.substring(startindex);
    }
    catch (Exception e) { }
    // if it fails, it's because we're out of range, so the empty string is OK
    
    return ret;
  }
  
  
  public char nextChar()
  {
  	// get rid of leading whitespace
  	// if you want the next char including whitespace,
  	// call nextAnyChar()
    lastLineRead = removeWhiteSpace(lastLineRead);
      
    if (lastLineRead.equals(""))
    {
        tokensWaiting = false;
    }
      
    while (!tokensWaiting) // from a previous read
    {
        lastLineRead = nextLine();  // refill the nextLineRead data member
        
        lastLineRead = removeWhiteSpace(lastLineRead);
      
        if (!lastLineRead.equals(""))
        {
            tokensWaiting = true;
        }
    
        if (good == false)
        {
            if (true /*theStream.nextToken() == StreamTokenizer.TT_EOF*/)
            {
                ; // do something here
                // perhaps print a warning and return some default value "0";
            }
        }  
    }
    
    // the next char must be non-whitespace, so we "remove" that
    // char from the lastLineRead member and return it
    char ret = lastLineRead.charAt(0);
    int linelen = lastLineRead.length();    
    lastLineRead = lastLineRead.substring(1,linelen);
    
    return ret;
  }
  
  public char nextAnyChar()
  {
  	// don't get rid of leading whitespace
  	// if you want the next char including whitespace,
  	// call nextAnyChar()
    // lastLineRead = removeWhiteSpace(lastLineRead);
    
    char ret = '!';
    
    if (lastLineRead.equals(""))
    {
    	// we return the newline char
        ret = '\n';
        
        // no tokens waiting -- important for nextLine to work correectly
        tokensWaiting = false;
        
        // now we need to get a new line for subsequent calls
        lastLineRead = nextLine();  // refill the nextLineRead data member
        
        // we don't remove whitespace in this case!
        //lastLineRead = removeWhiteSpace(lastLineRead);
      
        // nor do we check for empty lines!
        //if (!lastLineRead.equals(""))
        //{
        //    tokensWaiting = true;
        //}
    }
    else
    {
    	ret = lastLineRead.charAt(0);
	    int linelen = lastLineRead.length();    
	    lastLineRead = lastLineRead.substring(1,linelen);
    }
    
    return ret;
  }

  public String nextWord()
  {
    // get rid of leading whitespace
    lastLineRead = removeWhiteSpace(lastLineRead);
      
    if (lastLineRead.equals(""))
    {
        tokensWaiting = false;
    }
      
    while (!tokensWaiting) // from a previous read
    {
        lastLineRead = nextLine();  // refill the nextLineRead data member
        
        lastLineRead = removeWhiteSpace(lastLineRead);
      
        if (!lastLineRead.equals(""))
        {
            tokensWaiting = true;
        }
    
        if (good == false)
        {
            if (true /*theStream.nextToken() == StreamTokenizer.TT_EOF*/)
            {
                ; // do something here
                // perhaps print a warning and return some default value "0";
            }
        }  
    }
    
    // get chars up to the next white space
    int startindex = 0;
    int linelen = lastLineRead.length();
    
    while ( startindex<linelen && lastLineRead.charAt(startindex) > ' ' )
    {
        ++startindex;
    }
    
    String ret = "0";
    
    try {
        ret = lastLineRead.substring(0,startindex);
        lastLineRead = lastLineRead.substring(startindex,linelen);
    } catch (Exception e) { 
        System.err.println("error in nextWord! returning \"0\"");
        e.printStackTrace();
    };
    
    return ret;

  }
    
  public String nextLine()
  {
      String s = "";
      
      // get rid of leading whitespace, if there are tokens waiting
      
      if (tokensWaiting)
      {
          lastLineRead = removeWhiteSpace(lastLineRead);
          
          if (lastLineRead.equals(""))   // don't return an empty line
          {
            tokensWaiting = false;
          }
          else       // return the rest
          {
            tokensWaiting = false;
            return lastLineRead;
          }
      }
      
      if (!tokensWaiting) // from a previous read
      {
          try
          {
             s = r.readLine();
             tokensWaiting = true;
          }
          catch (Exception e)
          {
             good = false;
             // should we take this out?
             // I don't know when this even happens...
             // maybe if the stream is closed or something
             // but if it's a file that reaches the EOF,
             // s is null, but no exception is thrown
             System.err.println("There was a problem getting a line of input.");
             e.printStackTrace();
          }
      }
      
      // tokensWaiting should be false now, because this line is
      // all that's needed
      
      tokensWaiting = false;
      
      if (s == null)
      {
         // end of stream has been reached
         // check if we were reading from a file
         if (H.in != H.console_in)
         {
            // in this case, print a message and start taking
            // more input from stdin
            System.err.println("\n** The end of the input file has been reached.**");
            System.err.println(  "** Further input will be taken from the console:\n");
            H.inputFromConsole();
            // recursively call nextLine and return the result
            return H.in.nextLine(); 
         }
         else
         {
            // in this case, the input is already coming
            // from the console_in -- I don't know how s could
            // be null in this case, but perhaps we'll see...
            System.err.println("Attempt to read past the end of the stream/file.");
            System.err.println("returning the string 0");
            s = "0";
         }
      }
      
      // we check if we're echoing things to a file output
      // if so, we send the input out to the file
      // OR if the input is coming from a file, we echo it
      // either to another file or to the console
      //
      // so, the only time we _don't_ echo the input is when
      // it's coming from the stdin and not going to a file
      if (H.input_out != H.console_out || H.in != H.console_in)  // if file output
      {
         H.input_out.println(s);
      }
      
      // what if we're taking in the input from a file AND
      // sending it out to a file? It seems we should still
      // allow the user to see it! So we do...
      if (H.input_out != H.console_out && H.in != H.console_in) 
      {
         System.out.println(s);
      }
      return s;
      
  }
  
  // Thanks to Chris Erickson (HMC '06):
  
  public int nextInt()
  {
    return nextInt("An int please: ");
  }
  
  // allow a user-supplied String as a
  // prompt (repeated as often as necessary)
  
  public int nextInt(String e)
  {
    String s = "";
    int i = 0;
    boolean good = false;

    while ( !good )
    {
      s = nextWord();
      good = true;
      try
      {
        i = java.lang.Integer.parseInt(s);
      }
      catch ( NumberFormatException x )
      {
        good = false;
        System.out.print(e);
      }
    }

    return i;
  }
  
  // Thanks to Chris Erickson (HMC '06):
  
  public long nextLong()
  {
    return nextLong("A long int please: ");
  }
  
  // allow a user-supplied String as a
  // prompt (repeated as often as necessary)
  
  public long nextLong(String e)
  {
    String s = "";
    long i = 0l;
    boolean good = false;

    while ( !good )
    {
      s = nextWord();
      good = true;
      try
      {
        i = java.lang.Long.parseLong(s);
      }
      catch ( NumberFormatException x )
      {
        good = false;
        System.out.print(e);
      }
    }

    return i;
  }
  
  // Thanks to Chris Erickson (HMC '06):
  
  public float nextFloat()
  {
    return nextFloat("A float please: ");
  }
  
  // allow a user-supplied String as a
  // prompt (repeated as often as necessary)
  
  public float nextFloat(String e)
  {
    String s = "";
    float i = 0;
    boolean good = false;

    while ( !good )
    {
      s = nextWord();
      good = true;
      try
      {
        i = java.lang.Float.parseFloat(s);
      }
      catch ( NumberFormatException x )
      {
        good = false;
        System.out.print(e);
      }
    }

    return i;
  }

  // Thanks to Chris Erickson (HMC '06):
  
  public double nextDouble()
  {
    return nextDouble("A double please: ");
  }
  
  // allow a user-supplied String as a
  // prompt (repeated as often as necessary)
  
  public double nextDouble(String e)
  {
    String s = "";
    double i = 0;
    boolean good = false;

    while ( !good )
    {
      s = nextWord();
      good = true;
      try
      {
        i = java.lang.Double.parseDouble(s);
      }
      catch ( NumberFormatException x )
      {
        good = false;
        //System.out.print("**" + s + "**");
        System.out.print(e);
      }
    }

    return i;
  }
  

   


  static public void main(String[] args)
  {
    ;
  }
  
  
  
}


